from __future__ import annotations

from collections import UserDict
from contextvars import ContextVar, Token
from typing import final

from .base import UnionPattern
from .util import AllParam, Empty


@final
class Patterns(UserDict):
    def __init__(self, name):
        self.name = name
        super().__init__({"": Empty, "*": AllParam})

    def set(self, target, alias=None, cover=True):
        """
        增加可使用的类型转换器

        Args:
            target: 设置的表达式
            alias: 目标类型的别名
            cover: 是否覆盖已有的转换器
        """
        for k in {alias, target.alias, target.origin}:
            if not k:
                continue
            if k not in self.data or cover:
                self.data[k] = target
            else:
                al_pat = self.data[k]
                self.data[k] = (
                    UnionPattern([*al_pat.base, target])
                    if isinstance(al_pat, UnionPattern)
                    else (UnionPattern([al_pat, target]))
                )

    def sets(self, patterns, cover=True):
        for pat in patterns:
            self.set(pat, cover=cover)

    def merge(self, patterns):
        for k in patterns:
            self.set(patterns[k], alias=k)

    def remove(self, origin_type, alias=None):
        if alias and (al_pat := self.data.get(alias)):
            if isinstance(al_pat, UnionPattern):
                self.data[alias] = UnionPattern(filter(lambda x: x.alias != alias, al_pat.base))  # type: ignore
                if not self.data[alias].base:  # type: ignore # pragma: no cover
                    del self.data[alias]
            else:
                del self.data[alias]
        elif al_pat := self.data.get(origin_type):
            if isinstance(al_pat, UnionPattern):
                self.data[origin_type] = UnionPattern(
                    filter(lambda x: x.origin != origin_type, al_pat.for_validate)
                )
                if not self.data[origin_type].base:  # type: ignore # pragma: no cover
                    del self.data[origin_type]
            else:
                del self.data[origin_type]


_ctx = {"$global": Patterns("$global")}
_ctx_token: Token

pattern_ctx: ContextVar[Patterns] = ContextVar("nepatterns")
_ctx_token = pattern_ctx.set(_ctx["$global"])


def create_local_patterns(name, data=None, set_current=True) -> Patterns:
    """
    新建一个本地表达式组

    Args:
        name: 组名
        data: 可选的初始内容
        set_current: 是否设置为 current
    """
    global _ctx_token
    if name.startswith("$"):
        raise ValueError(name)
    new = Patterns(name)
    new.update(data or {})
    _ctx[name] = new
    if set_current:
        pattern_ctx.reset(_ctx_token)
        _ctx_token = pattern_ctx.set(new)
    return new


def switch_local_patterns(name):
    global _ctx_token
    if name.startswith("$"):
        raise ValueError(name)
    target = _ctx[name]
    pattern_ctx.reset(_ctx_token)
    _ctx_token = pattern_ctx.set(target)


def reset_local_patterns():
    global _ctx_token

    target = _ctx["$global"]
    pattern_ctx.reset(_ctx_token)
    _ctx_token = pattern_ctx.set(target)


def local_patterns():
    local = pattern_ctx.get()
    return local if local.name != "$global" else Patterns("$temp")


def global_patterns():
    return _ctx["$global"]


def all_patterns():
    """获取 global 与 local 的合并表达式组"""
    new = Patterns("$temp")
    local = local_patterns()
    if not local.name.startswith("$"):
        new.update(local_patterns().data)
    new.update(global_patterns().data)
    return new


__all__ = [
    "Patterns",
    "local_patterns",
    "global_patterns",
    "all_patterns",
    "switch_local_patterns",
    "create_local_patterns",
    "reset_local_patterns",
]



from __future__ import annotations

import contextlib
import dataclasses
import inspect
import sre_compile
import sys
import types
from typing import Any, List, Pattern, TypeVar, Literal, Union, get_args, get_origin, TYPE_CHECKING

from typing_extensions import Annotated

Empty = inspect.Signature.empty
if TYPE_CHECKING:
    from types import GenericAlias  # noqa
else:
    GenericAlias: type = type(List[int])
TPattern: type[Pattern] = type(sre_compile.compile("", 0))
AnnotatedType: type = type(Annotated[int, lambda x: x > 0])
Unions = (Union, types.UnionType) if sys.version_info >= (3, 10) else (Union,)  # pragma: no cover


class _All:
    """泛匹配"""

    def __repr__(self):
        return "AllParam"


AllParam = _All()


@dataclasses.dataclass
class RawStr:
    value: str


def generic_isinstance(obj: Any, par: type | Any | tuple[type, ...]) -> bool:
    """
    检查 obj 是否是 par 中的一个类型, 支持泛型, Any, Union, GenericAlias
    """
    if par is Any:
        return True
    with contextlib.suppress(TypeError):
        if isinstance(par, AnnotatedType):
            return generic_isinstance(obj, get_args(par)[0])
        if isinstance(par, type):
            return isinstance(obj, par)
        if get_origin(par) is Literal:
            return obj in get_args(par)
        if get_origin(par) in Unions:  # pragma: no cover
            return any(generic_isinstance(obj, p) for p in get_args(par))
        if isinstance(par, TypeVar):  # pragma: no cover
            if par.__constraints__:
                return any(generic_isinstance(obj, p) for p in par.__constraints__)
            return generic_isinstance(obj, par.__bound__) if par.__bound__ else True
        if isinstance(par, tuple):
            return any(generic_isinstance(obj, p) for p in par)
        if isinstance(obj, get_origin(par)):  # type: ignore
            return True
    return False



class Lang:
    type_error = "参数 {target} 的类型不正确"
    content_error = "参数 {target} 不正确"
    sequence_form_error = "不正确的序列形式: {target}"
    validate_reject = "{target} 校验失败"
    pattern_head_or_tail_error = "不允许正则表达式 {target} 头尾部分使用 '^' 或 '$' "


lang = Lang()



from __future__ import annotations

import re
from typing import Iterable, Any, Literal, TypeVar, Dict, Union

from .config import lang
from .core import BasePattern, MatchMode
from .exception import MatchFailed
from .util import Empty


class RegexPattern(BasePattern[Union[dict, tuple]]):
    """针对正则的特化匹配，支持正则组"""

    def __init__(self, pattern: str, alias: str | None = None):
        super().__init__(pattern, origin=Union[dict, tuple], alias=alias or 'regex[:group]')  # type: ignore

    def match(self, input_: str | Any):
        if not isinstance(input_, str):
            raise MatchFailed(lang.type_error.format(target=input_))
        if mat := self.regex_pattern.match(input_):
            return mat.groupdict() or mat.groups()
        raise MatchFailed(lang.content_error.format(target=input_))


class UnionPattern(BasePattern):
    """多类型参数的匹配"""

    optional: bool
    base: list[BasePattern | object | str]
    for_validate: list[BasePattern]
    for_equal: list[str | object]

    def __init__(self, base: Iterable[BasePattern | object | str], anti: bool = False):
        self.base = list(base)
        self.optional = False
        self.for_validate = []
        self.for_equal = []

        for arg in self.base:
            if arg == Empty:
                self.optional = True
                self.for_equal.append(None)
            elif isinstance(arg, BasePattern):
                self.for_validate.append(arg)
            else:
                self.for_equal.append(arg)
        alias_content = "|".join(
            [repr(a) for a in self.for_validate] + [repr(a) for a in self.for_equal]
        )
        super().__init__(
            r"(.+?)", MatchMode.KEEP, str, alias=alias_content, anti=anti
        )

    def match(self, text: str | Any):
        if not text:
            text = None
        if text not in self.for_equal:
            for pat in self.for_validate:
                if (res := pat.validate(text)).success:
                    return res.value
            raise MatchFailed(lang.content_error.format(target=text))
        return text

    def __repr__(self):
        return ("!" if self.anti else "") + (
            "|".join(repr(a) for a in (*self.for_validate, *self.for_equal))
        )

    def prefixed(self) -> UnionPattern:
        from .main import type_parser

        return UnionPattern(
            [pat.prefixed() for pat in self.for_validate]
            + [
                type_parser(eq).prefixed() if isinstance(eq, str) else eq  # type: ignore
                for eq in self.for_equal
            ],
            self.anti,
        )

    def suffixed(self) -> UnionPattern:
        from .main import type_parser

        return UnionPattern(
            [pat.suffixed() for pat in self.for_validate]
            + [
                type_parser(eq).suffixed() if isinstance(eq, str) else eq  # type: ignore
                for eq in self.for_equal
            ],
            self.anti,
        )


TSeq = TypeVar("TSeq", list, tuple, set)


class SequencePattern(BasePattern[TSeq]):
    """匹配列表或者元组或者集合"""

    base: BasePattern
    _mode: Literal["pre", "suf", "all"]

    def __init__(self, form: type[TSeq], base: BasePattern):
        self.base = base
        self._mode = "all"
        if form is list:
            super().__init__(
                r"\[(.+?)\]", MatchMode.REGEX_MATCH, form, alias=f"list[{base}]"
            )
        elif form is tuple:
            super().__init__(
                r"\((.+?)\)", MatchMode.REGEX_MATCH, form, alias=f"tuple[{base}]"
            )
        elif form is set:
            super().__init__(
                r"\{(.+?)\}", MatchMode.REGEX_MATCH, form, alias=f"set[{base}]"
            )
        else:
            raise ValueError(lang.sequence_form_error.format(target=str(form)))

    def match(self, text: str | Any):
        _res = super().match(text)
        _max = 0
        success: list[tuple[int, Any]] = []
        fail: list[tuple[int, MatchFailed]] = []
        for _max, s in enumerate(
                re.split(r"\s*,\s*", _res) if isinstance(_res, str) else _res
        ):
            try:
                success.append((_max, self.base.match(s)))
            except MatchFailed:
                fail.append((_max, MatchFailed(f"{s} is not matched with {self.base}")))

        if (
                (self._mode == "all" and fail)
                or (self._mode == "pre" and fail and fail[0][0] == 0)
                or (self._mode == "suf" and fail and fail[-1][0] == _max)
        ):
            raise fail[0][1]
        if self._mode == "pre" and fail:
            return self.origin(i[1] for i in success if i[0] < fail[0][0])
        if self._mode == "suf" and fail:
            return self.origin(i[1] for i in success if i[0] > fail[-1][0])
        return self.origin(i[1] for i in success)

    def __repr__(self):
        return f"{self.origin.__name__}[{self.base}]"

    def prefixed(self) -> SequencePattern:
        self._mode = "pre"
        return super(SequencePattern, self).prefixed()

    def suffixed(self) -> SequencePattern:
        self._mode = "suf"
        return super(SequencePattern, self).suffixed()


TKey = TypeVar("TKey")
TVal = TypeVar("TVal")


class MappingPattern(BasePattern[Dict[TKey, TVal]]):
    """匹配字典或者映射表"""

    key: BasePattern[TKey]
    value: BasePattern[TVal]
    _mode: Literal["pre", "suf", "all"]

    def __init__(self, arg_key: BasePattern[TKey], arg_value: BasePattern[TVal]):
        self.key = arg_key
        self.value = arg_value
        self._mode = "all"
        super().__init__(
            r"\{(.+?)\}",
            MatchMode.REGEX_MATCH,
            dict,
            alias=f"dict[{self.key}, {self.value}]",
        )

    def match(self, text: str | Any):
        _res = super().match(text)
        success: list[tuple[int, Any, Any]] = []
        fail: list[tuple[int, MatchFailed]] = []
        _max = 0

        def _generator_items(res: str | dict):
            if isinstance(res, dict):
                yield from res.items()
                return
            for m in re.split(r"\s*,\s*", res):
                yield re.split(r"\s*[:=]\s*", m)

        for _max, item in enumerate(_generator_items(_res)):
            k, v = item
            try:
                success.append((_max, self.key.match(k), self.value.match(v)))
            except MatchFailed:
                fail.append(
                    (
                        _max,
                        MatchFailed(
                            f"{k}: {v} is not matched with {self.key}: {self.value}"
                        ),
                    )
                )
        if (
                (self._mode == "all" and fail)
                or (self._mode == "pre" and fail and fail[0][0] == 0)
                or (self._mode == "suf" and fail and fail[-1][0] == _max)
        ):
            raise fail[0][1]
        if self._mode == "pre" and fail:
            return {i[1]: i[2] for i in success if i[0] < fail[0][0]}
        if self._mode == "suf" and fail:
            return {i[1]: i[2] for i in success if i[0] > fail[-1][0]}
        return {i[1]: i[2] for i in success}

    def __repr__(self):
        return f"dict[{self.key.origin.__name__}, {self.value}]"

    def prefixed(self) -> MappingPattern:
        self._mode = "pre"
        return super(MappingPattern, self).prefixed()

    def suffixed(self) -> MappingPattern:
        self._mode = "suf"
        return super(MappingPattern, self).suffixed()


_TCase = TypeVar("_TCase")


class SwitchPattern(BasePattern[_TCase]):
    switch: dict[Any, _TCase]

    def __init__(self, data: dict[Any | ellipsis, _TCase]):
        self.switch = data
        super().__init__("", MatchMode.TYPE_CONVERT, type(list(data.values())[0]))

    def __repr__(self):
        return "|".join(f"{k}" for k in self.switch if k != Ellipsis)

    def match(self, input_: Any) -> _TCase:
        try:
            return self.switch[input_]
        except KeyError as e:
            if Ellipsis in self.switch:
                return self.switch[...]
            raise MatchFailed(lang.content_error.format(target=input_)) from e


__all__ = ["RegexPattern", "UnionPattern", "SequencePattern", "MappingPattern", "SwitchPattern"]
